#!/usr/bin/env python-sirius

"""Test implementation.py."""

from unittest import TestCase

import siriuspy.util as util
import siriuspy.namesys as namesys


mock_flag = True

public_interface = (
    'get_siriuspvname_attrs',
    'join_name',
    'split_name',
    'get_pair_sprb',
    'SiriusPVName',
    'Filter',
)


class TestImplementation(TestCase):
    """Test Implementation module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            namesys, public_interface)
        self.assertTrue(valid)

    def test_get_siriuspvname_attrs(self):
        """Test  get_siriuspvname_attrs."""
        _attrs = [
            'channel_type',
            'prefix',
            'sec',
            'sub',
            'area_name',
            'dis',
            'dev',
            'idx',
            'device_name',
            'propty_name',
            'propty_suffix',
            'propty',
            'device_propty',
            'field',
        ]
        attrs = namesys.get_siriuspvname_attrs()
        self.assertEqual(sorted(_attrs), sorted(attrs))

    def test_join_name(self):
        """Test join_name."""
        # invalid arguments
        self.assertRaises(TypeError, namesys.join_name)
        self.assertRaises(TypeError, namesys.join_name,
                          sec='SI')
        self.assertRaises(TypeError, namesys.join_name,
                          sec='SI', sub='Fam', dis='PS')
        name = namesys.join_name(sec='SI', sub='Fam', dis='PS', dev='QDA')
        self.assertEqual(name, 'SI-Fam:PS-QDA')
        name = namesys.join_name(sec='BO', sub='Fam', dis='PS', dev='B',
                                 idx='1')
        self.assertEqual(name, 'BO-Fam:PS-B-1')
        name = namesys.join_name(sec='BO', sub='Fam', dis='PS', dev='B',
                                 idx='1', propty='PwrState-Sel')
        self.assertEqual(name, 'BO-Fam:PS-B-1:PwrState-Sel')
        name = namesys.join_name(sec='BO', sub='Fam', dis='PS', dev='B',
                                 idx='1', propty='Current-Mon', field='AVG')
        self.assertEqual(name, 'BO-Fam:PS-B-1:Current-Mon.AVG')

    def test_split_name(self):
        """Test split_name."""
        # empty string
        d = namesys.split_name('')
        self.assertEqual(len(d), len(namesys.get_siriuspvname_attrs()))
        for k, v in d.items():
            self.assertIsInstance(k, str)
            self.assertIsInstance(v, str)
            self.assertEqual(v, '')
        # Complete
        d = namesys.split_name(
            'ca://TEST-SI-Fam:PS-B1B2-1:PwrCurrent-SP.AVG')
        self.assertEqual(d['channel_type'], 'ca')
        self.assertEqual(d['prefix'], 'TEST')
        self.assertEqual(d['sec'], 'SI')
        self.assertEqual(d['sub'], 'Fam')
        self.assertEqual(d['dis'], 'PS')
        self.assertEqual(d['dev'], 'B1B2')
        self.assertEqual(d['idx'], '1')
        self.assertEqual(d['propty'], 'PwrCurrent-SP')
        self.assertEqual(d['field'], 'AVG')
        # one field
        self.assertRaises(IndexError, namesys.split_name, pvname='A:B')
        self.assertRaises(IndexError, namesys.split_name, pvname='A:B:C')
        self.assertRaises(IndexError, namesys.split_name, pvname='A:B-C:D')
        self.assertRaises(IndexError, namesys.split_name, pvname='A:B-C:D.F')


class TestSiriusPVName(TestCase):
    """Test SiriusPVName module."""

    public_interface = (
        'substitute',
        'device_name',
        'get_nickname',
        'is_write_pv',
        'is_sp_pv',
        'is_cmd_pv',
        'is_cte_pv',
        'is_rb_pv',
        'from_sp2rb',
        'from_rb2sp',
        'strip',
        'replace',
    )

    def test_public_interface(self):
        """Test SiriusPVName public interface."""
        valid = util.check_public_interface_namespace(
            namesys.SiriusPVName, TestSiriusPVName.public_interface)
        self.assertTrue(valid)

    def test_constructor(self):
        """Test constructor."""
        n = namesys.SiriusPVName('ca://PREFIX-SI-Fam:PS-B1B2-1:Current-SP.AVG')
        self.assertEqual(n.channel_type, 'ca')
        self.assertEqual(n.prefix, 'PREFIX')
        self.assertEqual(n.sec, 'SI')
        self.assertEqual(n.sub, 'Fam')
        self.assertEqual(n.dis, 'PS')
        self.assertEqual(n.dev, 'B1B2')
        self.assertEqual(n.idx, '1')
        self.assertEqual(n.device_name, 'SI-Fam:PS-B1B2-1')
        self.assertEqual(n.propty_name, 'Current')
        self.assertEqual(n.propty_suffix, 'SP')
        self.assertEqual(n.propty, 'Current-SP')
        self.assertEqual(n.device_propty, 'B1B2-1:Current-SP.AVG')
        self.assertEqual(n.field, 'AVG')

    def test_string(self):
        """Test string."""
        n = namesys.SiriusPVName('ca://PREFIX-SI-Fam:PS-B1B2-1:Current-SP.AVG')
        self.assertIsInstance(n, str)



"""AS RF configuration."""
from copy import deepcopy as _dcopy

# NOTE: absolute imports are necessary here due to how
# CONFIG_TYPES in __init__.py is built.
from siriuspy.clientconfigdb.types.global_config import _pvs_as_rf, \
    _pvs_li_llrf


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.

_pvs_bo_llrf = [
    ['BR-RF-DLLRF-01:ILK:REVSSA1:S', 0.0, 0.0],  # Interlock disable
    ['BR-RF-DLLRF-01:ILK:REVSSA2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:REVSSA3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:REVSSA4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:REVCAV:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:MAN:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:PLC:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:LLRF1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:LLRF2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:LLRF3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:PLG1:UP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:PLG1:DN:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:PLG2:UP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:PLG2:DN:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:VCAV:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:FWCAV:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:FWSSA1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN7:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN8:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN9:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN10:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN11:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN12:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN13:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN14:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:RFIN15:S', 0, 0.0],
    ['BR-RF-DLLRF-01:ILK:BEAM:TRIP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:SWITCHES:S', 0, 0.0],  # End switches logic
    ['BR-RF-DLLRF-01:TRIPINVERT:S', 0, 0.0],  # Beam trip logic
    ['BR-RF-DLLRF-01:VACINVERT:S', 0, 0.0],  # Vacuum sensor logic
    ['BR-RF-DLLRF-01:LIMIT:REVSSA1:S', 0, 0.0],  # Pwr intlck threshold [mV]
    ['BR-RF-DLLRF-01:LIMIT:REVSSA2:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:REVSSA3:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:REVSSA4:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:REVCAV:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:VCAV:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:FWCAV:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:FWSSA1:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN7:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN8:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN9:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN10:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN11:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN12:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN13:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN14:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:LIMIT:RFIN15:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:ILK:DELAY:S', 0, 0.0],  # [μs] Interlock delay
    ['BR-RF-DLLRF-01:mV:AL:REF-SP.DRVH', 0, 0.0],  # Settings PVs values lims
    ['BR-RF-DLLRF-01:mV:AL:REF-SP.DRVL', 0, 0.0],
    ['BR-RF-DLLRF-01:mV:RAMP:AMP:TOP-SP.DRVH', 0, 0.0],
    ['BR-RF-DLLRF-01:mV:RAMP:AMP:TOP-SP.DRVL', 0, 0.0],
    ['BR-RF-DLLRF-01:mV:RAMP:AMP:BOT-SP.DRVH', 0, 0.0],
    ['BR-RF-DLLRF-01:mV:RAMP:AMP:BOT-SP.DRVL', 0, 0.0],
    ['BR-RF-DLLRF-01:OLGAIN:S.DRVH', 0, 0.0],
    ['BR-RF-DLLRF-01:OLGAIN:S.DRVL', 0, 0.0],
    ['BR-RF-DLLRF-01:SL:KP:S.DRVH', 0, 0.0],  # kp limit high
    ['BR-RF-DLLRF-01:SL:KP:S.DRVL', 0, 0.0],  # kp limit low
    ['BO-05D:VA-CCG-RFC:FastRelay-SP', 0, 0.0],  # Pressure threshold
    ['BR-RF-DLLRF-01:AUTOCOND:S', 0, 0.0],  # Pressure Lock power increase
    ['BR-RF-DLLRF-01:EPS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FIM:S', 0, 0.0],
    ['BR-RF-DLLRF-01:PHSH:CAV:S', 0, 0.0],  # [°] # ADC Phase and Gain
    ['BR-RF-DLLRF-01:PHSH:FWDCAV:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:PHSH:FWDSSA1:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:PHSH:FWDSSA2:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:PHSH:FWDSSA3:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:PHSH:FWDSSA4:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:GAIN:FWDCAV:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:FWDSSA1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:FWDSSA2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:FWDSSA3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:FWDSSA4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:PHSH:SSA1:S', 0, 0.0],  # [°] # DAC Phse and Gain
    ['BR-RF-DLLRF-01:PHSH:SSA2:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:PHSH:SSA3:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:PHSH:SSA4:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:GAIN:SSA1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:SSA2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:SSA3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:GAIN:SSA4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:SL:KP:S', 0, 0.0],  # Loops parameters
    ['BR-RF-DLLRF-01:SL:KI:S', 0, 0.0],
    ['BR-RF-DLLRF-01:SL:PILIMIT:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:SL:SEL:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FL:KP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FL:KI:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FL:PILIMIT:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:FL:SEL:S', 0, 0.0],
    ['BR-RF-DLLRF-01:AL:KP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:AL:KI:S', 0, 0.0],
    ['BR-RF-DLLRF-01:AL:SEL:S', 0, 0.0],
    ['BR-RF-DLLRF-01:PL:KP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:PL:KI:S', 0, 0.0],
    ['BR-RF-DLLRF-01:PL:SEL:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MODE:S', 0, 0.0],  # Loop mode
    ['BR-RF-DLLRF-01:FWMIN:AMPPHS:S', 0, 0.0],  # [mV] # Min forward power
    ['BR-RF-DLLRF-01:mV:AMPREF:MIN:S', 0, 0.0],  # [mV] # Min amplitude ref.
    ['BR-RF-DLLRF-01:PHSREF:MIN:S', 0, 0.0],  # [°] # Min phase ref.
    ['BR-RF-DLLRF-01:OLGAIN:S', 0, 0.0],  # Open loop gain
    ['BR-RF-DLLRF-01:TUNE:POS:S', 0, 0.0],  # Tuning loop config
    ['BR-RF-DLLRF-01:TUNE:FWMIN:S', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:TUNE:MARGIN:HI:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:TUNE:MARGIN:LO:S', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:DTune-SP', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:TUNE:DELAY:S', 0, 0.0],  # [s]
    ['BR-RF-DLLRF-01:TUNE:PULSE:FREQ:S', 0, 0.0],
    ['BR-RF-DLLRF-01:TUNE:FILT:S', 0, 0.0],
    ['BR-RF-DLLRF-01:TUNE:TRIG:S', 0, 0.0],
    ['BR-RF-DLLRF-01:TUNE:TOPRAMP:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FF:POS:S', 0, 0.0],  # Field Flatness loop config
    ['BR-RF-DLLRF-01:FF:DEADBAND:S', 0, 0.0],  # [%]
    ['BR-RF-DLLRF-01:FF:GAIN:CELL2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FF:GAIN:CELL4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:freq:cond:S', 0, 0.0],  # [Hz] # Pulsed mode config
    ['BR-RF-DLLRF-01:freq:duty:S', 0, 0.0],  # [%]
    ['BR-RF-DLLRF-01:RmpTs1-SP', 0, 0.0],  # [ms] # Ramp mode config
    ['BR-RF-DLLRF-01:RmpTs2-SP', 0, 0.0],  # [ms]
    ['BR-RF-DLLRF-01:RmpTs3-SP', 0, 0.0],  # [ms]
    ['BR-RF-DLLRF-01:RmpTs4-SP', 0, 0.0],  # [ms]
    ['BR-RF-DLLRF-01:mV:RAMP:AMP:BOT-SP', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:RmpPhsBot-SP', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:mV:RAMP:AMP:TOP-SP', 0, 0.0],  # [mV]
    ['BR-RF-DLLRF-01:RmpPhsTop-SP', 0, 0.0],  # [°]
    ['BR-RF-DLLRF-01:RmpIncTs-SP', 0, 0.0],  # [min]
    ['BR-RF-DLLRF-01:DisableRampDown:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FDL:FrameQty-SP', 0, 0.0],
    ['BR-RF-DLLRF-01:FDL:REARM', 0, 0.0],
    ['RA-RaBO01:RF-CavPlDrivers:DrEnbl-Sel', 0, 0.0],  # Enable plungers step
                                                       # motor drivers
    ['BR-RF-DLLRF-01:PHSH:ADC:S', 0, 0.0],  # Enable ADC phase and gain
    ['BR-RF-DLLRF-01:PHSH:DAC:S', 0, 0.0],  # Enable DAC phase and gain
    ]


_pvs_bo_rfssa = [
    ['RA-ToBO:OffsetConfig:UpperIncidentPower', 0, 0.0],  # [dB] Offsets
                                                          # SSA Tower
    ['RA-ToBO:OffsetConfig:UpperReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:LowerIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:LowerReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:InputIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:InputReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:OutputIncidentPower', 0, 0.0],  # [dB]
    ['RA-ToBO:OffsetConfig:OutputReflectedPower', 0, 0.0],  # [dB]
    ['RA-ToBO:AlarmConfig:GeneralPowerLimHiHi', 0, 0.0],  # [dBm] Alarms limits
                                                          # power SSA tower
    ['RA-ToBO:AlarmConfig:GeneralPowerLimHigh', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:GeneralPowerLimLow', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:GeneralPowerLimLoLo', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimHigh', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimLow', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],  # [dBm]
    ['RA-ToBO:AlarmConfig:CurrentLimHiHi', 0, 0.0],  # [A] Alarms limits
                                                     # currents SSA tower
    ['RA-ToBO:AlarmConfig:CurrentLimHigh', 0, 0.0],  # [A]
    ['RA-ToBO:AlarmConfig:CurrentLimLow', 0, 0.0],  # [A]
    ['RA-ToBO:AlarmConfig:CurrentLimLoLo', 0, 0.0],  # [A]
    ]


_pvs_bo_rfcal = [
    ['BR-RF-DLLRF-01:CAV:Const:OFS:S', 0, 0.0],  # Offsets and conv coeffs
    ['BR-RF-DLLRF-01:CAV:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:U-Raw:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:U-Raw:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:U-Raw:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:U-Raw:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CAV:Const:U-Raw:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCAV:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCAV:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCAV:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCAV:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCAV:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCAV:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MO:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MO:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MO:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MO:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MO:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:MO:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVSSA1:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL2:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL2:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL2:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL2:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL2:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL2:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL4:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL4:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL4:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL4:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL4:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL4:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL1:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL1:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL1:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL1:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL1:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL1:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL5:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL5:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL5:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL5:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL5:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:CELL5:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:INPRE:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:INPRE:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:INPRE:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:INPRE:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:INPRE:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:INPRE:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDPRE:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDPRE:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDPRE:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDPRE:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDPRE:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDPRE:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVPRE:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVPRE:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVPRE:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVPRE:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVPRE:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVPRE:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCIRC:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCIRC:Const:OFS:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCIRC:Const:Raw-U:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCIRC:Const:Raw-U:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCIRC:Const:Raw-U:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCIRC:Const:Raw-U:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:REVCIRC:Const:Raw-U:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:CAV:Const:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:CAV:Const:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:CAV:Const:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:CAV:Const:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:CAV:Const:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDCAV:Const:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDCAV:Const:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDCAV:Const:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDCAV:Const:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDCAV:Const:C4:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDSSA1:Const:C0:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDSSA1:Const:C1:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDSSA1:Const:C2:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDSSA1:Const:C3:S', 0, 0.0],
    ['BR-RF-DLLRF-01:OLG:FWDSSA1:Const:C4:S', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:AmpVCav2HwCoeff0-SP', 0, 0.0],  # Conv coeffs for
                                                        # Voltage Gap calc
    ['RA-RaBO01:RF-LLRF:AmpVCav2HwCoeff1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:AmpVCav2HwCoeff2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:AmpVCav2HwCoeff3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:AmpVCav2HwCoeff4-SP', 0, 0.0],
    ['BO-05D:RF-P5Cav:Rsh-SP', 0, 0.0],  # Cavity Shunt impedance
    ['RA-RaBO01:RF-LLRF:Hw2AmpVCavCoeff0-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Hw2AmpVCavCoeff1-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Hw2AmpVCavCoeff2-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Hw2AmpVCavCoeff3-SP', 0, 0.0],
    ['RA-RaBO01:RF-LLRF:Hw2AmpVCavCoeff4-SP', 0, 0.0],
    ]


_pvs_si_llrf = [
    ['SR-RF-DLLRF-01:ILK:REVSSA1:S', 0, 0.0],  # Interlock disable
    ['SR-RF-DLLRF-01:ILK:REVSSA2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:REVSSA3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:REVSSA4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:REVCAV:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:MAN:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:PLC:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:LLRF1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:LLRF2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:LLRF3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:PLG1:UP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:PLG1:DN:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:PLG2:UP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:PLG2:DN:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:VCAV:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:FWCAV:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:FWSSA1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN7:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN8:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN9:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN10:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN11:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN12:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN13:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN14:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:RFIN15:S', 0, 0.0],
    ['SR-RF-DLLRF-01:ILK:BEAM:TRIP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:SWITCHES:S', 0, 0.0],  # End switches logic
    ['SR-RF-DLLRF-01:TRIPINVERT:S', 0, 0.0],  # Beam trip logic
    ['SR-RF-DLLRF-01:VACINVERT:S', 0, 0.0],  # Vacuum sensor logic
    ['SR-RF-DLLRF-01:LIMIT:REVSSA1:S', 0, 0.0],  # [mV] Pwr intlck threshold
    ['SR-RF-DLLRF-01:LIMIT:REVSSA2:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:REVSSA3:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:REVSSA4:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:REVCAV:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:VCAV:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:FWCAV:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:FWSSA1:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN7:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN8:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN9:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN10:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN11:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN12:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN13:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN14:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:LIMIT:RFIN15:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:ILK:DELAY:S', 0, 0.0],  # [μs] # Interlock delay
    ['SR-RF-DLLRF-01:mV:AL:REF-SP.DRVH', 0, 0.0],  # [mV] Set PVs value lims
    ['SR-RF-DLLRF-01:mV:AL:REF-SP.DRVL', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:OLGAIN:S.DRVH', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:OLGAIN:S.DRVL', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:SL:KP:S.DRVH', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:SL:KP:S.DRVL', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:AUTOCOND:S', 0, 0.0],  # Pressure Lock power increase
    ['SR-RF-DLLRF-01:EPS:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FIM:S', 0, 0.0],
    ['SR-RF-DLLRF-01:PHSH:CAV:S', 0, 0.0],  # [°] ADC Phase and Gain
    ['SR-RF-DLLRF-01:PHSH:FWDCAV:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:PHSH:FWDSSA1:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:PHSH:FWDSSA2:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:PHSH:FWDSSA3:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:PHSH:FWDSSA4:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:GAIN:FWDCAV:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:FWDSSA1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:FWDSSA2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:FWDSSA3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:FWDSSA4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:PHSH:SSA1:S', 0, 0.0],  # [°] DAC Phse and Gain
    ['SR-RF-DLLRF-01:PHSH:SSA2:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:PHSH:SSA3:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:PHSH:SSA4:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:GAIN:SSA1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:SSA2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:SSA3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:GAIN:SSA4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:SL:KP:S', 0, 0.0],  # Loops parameters
    ['SR-RF-DLLRF-01:SL:KI:S', 0, 0.0],
    ['SR-RF-DLLRF-01:SL:PILIMIT:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:SL:SEL:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FL:KP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FL:KI:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FL:PILIMIT:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:FL:SEL:S', 0, 0.0],
    ['SR-RF-DLLRF-01:AL:KP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:AL:KI:S', 0, 0.0],
    ['SR-RF-DLLRF-01:AL:SEL:S', 0, 0.0],
    ['SR-RF-DLLRF-01:PL:KP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:PL:KI:S', 0, 0.0],
    ['SR-RF-DLLRF-01:PL:SEL:S', 0, 0.0],
    ['SR-RF-DLLRF-01:MODE:S', 0, 0.0],  # Loop mode
    ['SR-RF-DLLRF-01:FWMIN:AMPPHS:S', 0, 0.0],  # [mV] Min forward power
    ['SR-RF-DLLRF-01:mV:AMPREF:MIN:S', 0, 0.0],  # [mV] Min amplitude reference
    ['SR-RF-DLLRF-01:PHSREF:MIN:S', 0, 0.0],  # [°] Min phase reference
    ['SR-RF-DLLRF-01:OLGAIN:S', 0, 0.0],  # Open loop gain
    ['SR-RF-DLLRF-01:PL:REF:S', 0, 0.0],  # [°] Phase ref
    ['SR-RF-DLLRF-01:TUNE:POS:S', 0, 0.0],  # Tuning loop config
    ['SR-RF-DLLRF-01:TUNE:FWMIN:S', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:TUNE:MARGIN:HI:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:TUNE:MARGIN:LO:S', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:DTune-SP', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:TUNE:DELAY:S', 0, 0.0],  # [s]
    ['SR-RF-DLLRF-01:TUNE:PULSE:FREQ:S', 0, 0.0],
    ['SR-RF-DLLRF-01:TUNE:FILT:S', 0, 0.0],
    ['SR-RF-DLLRF-01:TUNE:TRIG:S', 0, 0.0],
    ['SR-RF-DLLRF-01:TUNE:TOPRAMP:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FF:POS:S', 0, 0.0],  # Field Flatness loop config
    ['SR-RF-DLLRF-01:FF:DEADBAND:S', 0, 0.0],  # [%]
    ['SR-RF-DLLRF-01:FF:GAIN:CELL2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FF:GAIN:CELL4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:COND:DC:S', 0, 0.0],  # [%]
    ['SR-RF-DLLRF-01:RmpTs1-SP', 0, 0.0],  # [ms] Ramp mode config
    ['SR-RF-DLLRF-01:RmpTs2-SP', 0, 0.0],  # [ms]
    ['SR-RF-DLLRF-01:RmpTs3-SP', 0, 0.0],  # [ms]
    ['SR-RF-DLLRF-01:RmpTs4-SP', 0, 0.0],  # [ms]
    ['SR-RF-DLLRF-01:mV:RAMP:AMP:TOP-SP', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:RmpPhsTop-SP', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:mV:RAMP:AMP:BOT-SP', 0, 0.0],  # [mV]
    ['SR-RF-DLLRF-01:RmpPhsBot-SP', 0, 0.0],  # [°]
    ['SR-RF-DLLRF-01:RmpIncTs-SP', 0, 0.0],  # [min]
    ['SR-RF-DLLRF-01:DisableRampDown:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FDL:FrameQty-SP', 0, 0.0],
    ['SR-RF-DLLRF-01:FDL:REARM', 0, 0.0],
    ['SR-RF-DLLRF-01:PHSH:ADC:S', 0, 0.0],  # Enable ADC phase and gain
    ['SR-RF-DLLRF-01:PHSH:DAC:S', 0, 0.0],  # Enable DAC phase and gain
    ]


_pvs_si_rfssa = [
    ['RA-ToSIA01:OffsetConfig:UpperIncidentPower', 0, 0.0],  # Offsets SSA Twr1
    ['RA-ToSIA01:OffsetConfig:UpperReflectedPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:LowerIncidentPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:LowerReflectedPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:InputIncidentPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:InputReflectedPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:OutputIncidentPower', 0, 0.0],
    ['RA-ToSIA01:OffsetConfig:OutputReflectedPower', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:GeneralPowerLimHiHi', 0, 0.0],  # Alarms lims pwr
                                                             # SSA tower 1
    ['RA-ToSIA01:AlarmConfig:GeneralPowerLimHigh', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:GeneralPowerLimLow', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:GeneralPowerLimLoLo', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:CurrentLimHiHi', 0, 0.0],  # Alarms lims currents
                                                        # SSA tower 1
    ['RA-ToSIA01:AlarmConfig:CurrentLimHigh', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:CurrentLimLow', 0, 0.0],
    ['RA-ToSIA01:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:UpperIncidentPower', 0, 0.0],  # Offsets SSA Twr2
    ['RA-ToSIA02:OffsetConfig:UpperReflectedPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:LowerIncidentPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:LowerReflectedPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:InputIncidentPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:InputReflectedPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:OutputIncidentPower', 0, 0.0],
    ['RA-ToSIA02:OffsetConfig:OutputReflectedPower', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:GeneralPowerLimHiHi', 0, 0.0],  # Alarms lims pwr
                                                             # SSA tower 2
    ['RA-ToSIA02:AlarmConfig:GeneralPowerLimHigh', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:GeneralPowerLimLow', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:GeneralPowerLimLoLo', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimHiHi', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimHigh', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimLow', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:InnerPowerLimLoLo', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:CurrentLimHiHi', 0, 0.0],  # Alarms lims currents
                                                        # SSA tower 2
    ['RA-ToSIA02:AlarmConfig:CurrentLimHigh', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:CurrentLimLow', 0, 0.0],
    ['RA-ToSIA02:AlarmConfig:CurrentLimLoLo', 0, 0.0],
    ]


_pvs_si_rfcav = [
    ['SI-02SB:RF-P7Cav:Disc1FlwRt-Mon', 0, 0.0],  # [L/h] CavP7 water flowrate
    ['SI-02SB:RF-P7Cav:Cell1FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc2FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Cell2FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc3FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Cell3FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc4FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Cell4FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc5FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Cell5FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc6FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Cell6FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc7FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Cell7FlwRt-Mon', 0, 0.0],  # [L/h]
    ['SI-02SB:RF-P7Cav:Disc8FlwRt-Mon', 0, 0.0],  # [L/h]
    ]


_pvs_si_rfcal = [
    ['SR-RF-DLLRF-01:CAV:Const:OFS:S', 0, 0.0],  # [dB] Offsets and conv coeffs
    ['SR-RF-DLLRF-01:CAV:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:U-Raw:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:U-Raw:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:U-Raw:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:U-Raw:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CAV:Const:U-Raw:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCAV:Const:U-Raw:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVCAV:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:REVCAV:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVCAV:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVCAV:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVCAV:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVCAV:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:MO:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:MO:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:MO:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:MO:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:MO:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:MO:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA1:Const:U-Raw:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA1:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA1:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL2:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:CELL2:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL2:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL2:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL2:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL2:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL6:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:CELL6:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL6:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL6:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL6:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:CELL6:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA2:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:FWDSSA2:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA2:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA2:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA2:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDSSA2:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA2:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:REVSSA2:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA2:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA2:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA2:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:REVSSA2:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE1:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:INPRE1:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE1:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE1:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE1:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE1:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE1:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:FWDPRE1:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE1:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE1:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE1:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE1:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE2:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:INPRE2:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE2:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE2:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE2:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:INPRE2:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE2:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:FWDPRE2:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE2:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE2:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE2:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDPRE2:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCIRC:Const:OFS:S', 0, 0.0],  # [dB]
    ['SR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:FWDCIRC:Const:Raw-U:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:CAV:Const:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:CAV:Const:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:CAV:Const:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:CAV:Const:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:CAV:Const:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDCAV:Const:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDCAV:Const:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDCAV:Const:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDCAV:Const:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDCAV:Const:C4:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDSSA1:Const:C0:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDSSA1:Const:C1:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDSSA1:Const:C2:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDSSA1:Const:C3:S', 0, 0.0],
    ['SR-RF-DLLRF-01:OLG:FWDSSA1:Const:C4:S', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:AmpVCav2HwCoeff0-SP', 0, 0.0],  # Conv coeffs for
                                                         # Voltage Gap calc
    ['RA-RaSIA01:RF-LLRF:AmpVCav2HwCoeff1-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:AmpVCav2HwCoeff2-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:AmpVCav2HwCoeff3-SP', 0, 0.0],
    ['RA-RaSIA01:RF-LLRF:AmpVCav2HwCoeff4-SP', 0, 0.0],
    ['SI-02SB:RF-P7Cav:Rsh-SP', 0, 0.0],  # [Ohm] # Cavity Shunt impedance
    ]


_template_dict = {
    'pvs':
    _pvs_as_rf +
    _pvs_li_llrf + _pvs_bo_llrf + _pvs_bo_rfssa + _pvs_bo_rfcal +
    _pvs_si_llrf + _pvs_si_rfssa + _pvs_si_rfcav + _pvs_si_rfcal
    }



#!/usr/local/bin/python-sirius
"""LI Diag PVs."""

import numpy as _np

from ...util import update_bit as _update_bit
from .csdev import Const as _Const


class LIScalarDiffPV:
    """Diff of a LI setpoint and a readback PVs."""

    SP = 0
    RB = 1

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute difference between SP and RB values."""
        disconnected = \
            not computed_pv.pvs[LIScalarDiffPV.SP].connected or \
            not computed_pv.pvs[LIScalarDiffPV.RB].connected
        if disconnected:
            return None
        value_sp = computed_pv.pvs[LIScalarDiffPV.SP].value
        value_rb = computed_pv.pvs[LIScalarDiffPV.RB].value
        diff = value_rb - value_sp
        return {'value': diff}


class LIVecDiffPV:
    """Diff of a LI setpoint and a readback PVs."""

    I_SETT = 0
    Q_SETT = 1
    I_DATA = 2
    Q_DATA = 3

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute difference between SP and RB values."""
        disconnected = \
            not computed_pv.pvs[LIVecDiffPV.I_SETT].connected or \
            not computed_pv.pvs[LIVecDiffPV.Q_SETT].connected or \
            not computed_pv.pvs[LIVecDiffPV.I_DATA].connected or \
            not computed_pv.pvs[LIVecDiffPV.Q_DATA].connected
        if disconnected:
            return None
        sp_i = computed_pv.pvs[LIVecDiffPV.I_SETT].value
        sp_q = computed_pv.pvs[LIVecDiffPV.Q_SETT].value
        sp_vec = _np.array([sp_i, sp_q])
        rb_i = computed_pv.pvs[LIVecDiffPV.I_DATA].value
        rb_q = computed_pv.pvs[LIVecDiffPV.Q_DATA].value
        rb_vec = _np.array([rb_i, rb_q])
        diff = _np.linalg.norm(rb_vec - sp_vec)
        return {'value': diff}


class LIRFStatusPV:
    """LI RF Status PV."""

    BIT_DISCONN = 0b00000001
    BIT_STATOFF = 0b00000010
    BIT_TRIGOFF = 0b00000100
    BIT_INTGOFF = 0b00001000
    BIT_FBMDOFF = 0b00010000
    BIT_AMPDIFF = 0b00100000
    BIT_PHSDIFF = 0b01000000
    BIT_IXQDIFF = 0b10000000

    PV_STREAM = 0
    PV_EXTTRG = 1
    PV_INTEGR = 2
    PV_FBMODE = 3
    PV_AMPDIF = 4
    PV_PHSDIF = 5
    PV_IXQDIF = 6

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute Status PV."""
        value = 0

        # connected?
        disconnected = \
            not computed_pv.pvs[LIRFStatusPV.PV_STREAM].connected or \
            not computed_pv.pvs[LIRFStatusPV.PV_EXTTRG].connected or \
            not computed_pv.pvs[LIRFStatusPV.PV_INTEGR].connected or \
            not computed_pv.pvs[LIRFStatusPV.PV_FBMODE].connected or \
            not computed_pv.pvs[LIRFStatusPV.PV_AMPDIF].connected or \
            not computed_pv.pvs[LIRFStatusPV.PV_PHSDIF].connected or \
            not computed_pv.pvs[LIRFStatusPV.PV_IXQDIF].connected
        if disconnected:
            value |= LIRFStatusPV.BIT_DISCONN
            value |= LIRFStatusPV.BIT_STATOFF
            value |= LIRFStatusPV.BIT_TRIGOFF
            value |= LIRFStatusPV.BIT_INTGOFF
            value |= LIRFStatusPV.BIT_FBMDOFF
            value |= LIRFStatusPV.BIT_AMPDIFF
            value |= LIRFStatusPV.BIT_PHSDIFF
            value |= LIRFStatusPV.BIT_IXQDIFF
            return {'value': value}

        # state?
        sts = computed_pv.pvs[LIRFStatusPV.PV_STREAM].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIRFStatusPV.BIT_STATOFF

        # trigger?
        sts = computed_pv.pvs[LIRFStatusPV.PV_EXTTRG].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIRFStatusPV.BIT_TRIGOFF

        # integral?
        sts = computed_pv.pvs[LIRFStatusPV.PV_INTEGR].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIRFStatusPV.BIT_INTGOFF

        # feedback?
        sts = computed_pv.pvs[LIRFStatusPV.PV_FBMODE].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIRFStatusPV.BIT_FBMDOFF

        # amp diff?
        severity = computed_pv.pvs[LIRFStatusPV.PV_AMPDIF].severity
        if severity != 0:
            value |= LIRFStatusPV.BIT_AMPDIFF

        # phase diff?
        severity = computed_pv.pvs[LIRFStatusPV.PV_PHSDIF].severity
        if severity != 0:
            value |= LIRFStatusPV.BIT_PHSDIFF

        # IxQ diff?
        severity = computed_pv.pvs[LIRFStatusPV.PV_IXQDIF].severity
        if severity != 0:
            value |= LIRFStatusPV.BIT_IXQDIFF

        return {'value': value}


class LIPUStatusPV:
    """LI PU Status PV."""

    BIT = _Const.register(
        'BIT',
        ['DISCONN', 'RUNSTOP', 'PREHEAT', 'CHRALLW', 'TRGALLW',
         'EMRSTOP', 'CPS_ALL', 'THYHEAT', 'KLYHEAT', 'LVRDYOK',
         'HVRDYOK', 'TRRDYOK', 'SELFFLT', 'SYS_RDY', 'TRGNORM',
         'PLSCURR', 'VLTDIFF', 'CRRDIFF', 'CHARGES', 'TRIGOUT'])

    PV = _Const.register(
        'PV',
        ['CHARGES', 'TRIGOUT', 'RUNSTOP', 'PREHEAT', 'CHRALLW',
         'TRGALLW', 'EMRSTOP', 'CPS_ALL', 'THYHEAT', 'KLYHEAT',
         'LVRDYOK', 'HVRDYOK', 'TRRDYOK', 'SELFFLT', 'SYS_RDY',
         'TRGNORM', 'PLSCURR', 'VLTDIFF', 'CRRDIFF'])

    PVS_OFF_IGNORE = [
        'CHRALLW', 'TRGALLW', 'CPS_ALL', 'TRRDYOK', 'SELFFLT',
        'SYS_RDY', 'TRGNORM', 'PLSCURR']

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute Status PV."""
        value = 0

        # connected?
        disconnected = \
            any([not computed_pv.pvs[i].connected for i in LIPUStatusPV.PV])
        if disconnected:
            value = (1 << len(LIPUStatusPV.BIT)+1)-1
            return {'value': value}

        chrgsts = computed_pv.pvs[LIPUStatusPV.PV.CHARGES].value
        trigsts = computed_pv.pvs[LIPUStatusPV.PV.TRIGOUT].value
        for pvi, attr in enumerate(LIPUStatusPV.PV._fields):
            bit = getattr(LIPUStatusPV.BIT, attr)
            if 'DIFF' in attr:
                sts = computed_pv.pvs[pvi].severity
                value = _update_bit(value, bit, sts != 0)
            else:
                if (not chrgsts or not trigsts) and \
                        attr in LIPUStatusPV.PVS_OFF_IGNORE:
                    continue
                sts = computed_pv.pvs[pvi].value
                value = _update_bit(value, bit, sts != 1 or sts is None)

        return {'value': value}


class LIEGHVStatusPV:
    """LI Egun HVPS Status PV."""

    BIT_DISCONN = 0b0001
    BIT_SWITOFF = 0b0010
    BIT_ENBLOFF = 0b0100
    BIT_VOLTDIF = 0b1000

    PV_SWITCH = 0
    PV_ENABLE = 1
    PV_VLTDIF = 2

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute Status PV."""
        value = 0

        # connected?
        disconnected = \
            not computed_pv.pvs[LIEGHVStatusPV.PV_SWITCH].connected or \
            not computed_pv.pvs[LIEGHVStatusPV.PV_ENABLE].connected or \
            not computed_pv.pvs[LIEGHVStatusPV.PV_VLTDIF].connected
        if disconnected:
            value |= LIEGHVStatusPV.BIT_DISCONN
            value |= LIEGHVStatusPV.BIT_SWITOFF
            value |= LIEGHVStatusPV.BIT_ENBLOFF
            value |= LIEGHVStatusPV.BIT_VOLTDIF
            return {'value': value}

        # switch?
        sts = computed_pv.pvs[LIEGHVStatusPV.PV_SWITCH].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIEGHVStatusPV.BIT_SWITOFF

        # enable?
        sts = computed_pv.pvs[LIEGHVStatusPV.PV_ENABLE].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIEGHVStatusPV.BIT_ENBLOFF

        # volt diff?
        severity = computed_pv.pvs[LIEGHVStatusPV.PV_VLTDIF].severity
        if severity != 0:
            value |= LIEGHVStatusPV.BIT_VOLTDIF

        return {'value': value}


class LIFilaPSStatusPV:
    """LI FilaPS Status PV."""

    BIT_DISCONN = 0b0001
    BIT_SWITOFF = 0b0010
    BIT_CURRDIF = 0b0100

    PV_SWITCH = 0
    PV_CURDIF = 1

    def compute_update(self, computed_pv, updated_pv_name, value):
        """Compute Status PV."""
        value = 0

        # connected?
        disconnected = \
            not computed_pv.pvs[LIFilaPSStatusPV.PV_SWITCH].connected or \
            not computed_pv.pvs[LIFilaPSStatusPV.PV_CURDIF].connected
        if disconnected:
            value |= LIFilaPSStatusPV.BIT_DISCONN
            value |= LIFilaPSStatusPV.BIT_SWITOFF
            value |= LIFilaPSStatusPV.BIT_CURRDIF
            return {'value': value}

        # switch?
        sts = computed_pv.pvs[LIFilaPSStatusPV.PV_SWITCH].value
        if sts != _Const.OffOn.On or sts is None:
            value |= LIFilaPSStatusPV.BIT_SWITOFF

        # volt diff?
        severity = computed_pv.pvs[LIFilaPSStatusPV.PV_CURDIF].severity
        if severity != 0:
            value |= LIFilaPSStatusPV.BIT_CURRDIF

        return {'value': value}



"""TS posang correction configurations.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return a dict with ramp settings."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }

    return _dict


# Position-Angle Correction Response Matrices for the Storage Ring Injection
# (TS correctors)
#
# Horizontal Matrix:
#
# | DeltaPosX @TS-04:DI-Scrn-3 |   | h11 h12 |   | Kick TS-04:PU-InjSeptG(1&2)|
# |                            | = |         | * |       or TS-04:PS-CH       |
# | DeltaAngX @TS-04:DI-Scrn-3 |   | h21 h22 |   |   Kick TS-04:PU-InjSeptF   |
#
# Data structure:
#         h11   h12
#         h21   h22
#
# Vertical Matrix:
#
# | DeltaPosY @TS-04:DI-Scrn-3 |   | v11 v12 v13 v14 |   | Kick TS-04:PS-CV-0 |
# |                            | = |                 | * |                    |
# | DeltaAngY @TS-04:DI-Scrn-3 |   | v21 v22 v23 v24 |   | Kick TS-04:PS-CV-1 |
#                                                        |                    |
#                                                        |Kick TS-04:PS-CV-1E2|
#                                                        |                    |
#                                                        | Kick TS-04:PS-CV-2 |
#
# Data structure:
#         v11   v12
#         v21   v22

_template_dict = {
    'respm-x': [[0.0, 0.0],
                [0.0, 0.0]],
    'respm-y': [[0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0]],
}



"""Magnet utils."""

import math as _math
import re as _re

import numpy as _numpy


from ..namesys import SiriusPVName as _SiriusPVName


def get_nominal_dipole_angles():
    """Return individual nominal angles [rad] of all Sirius dipole types."""
    ref_angles = {
        'SI_BC': _math.radians(4.296600),
        'SI_B1': _math.radians(2.755300),
        'SI_B2': _math.radians(4.096400),
        'TS': _math.radians(5.011542),  # TS.V03.03
        'BO': _math.radians(7.200000),
        'TB': _math.radians(15.00000),
        'LI': _math.radians(-45.00000),
    }
    return ref_angles


def get_magfunc_2_multipole_dict():
    """Return multipole dict given the magnetic function.

    Conventions:
        1. harmonics: 0 (dipole), 1 (quadrupole), 2 (sextupole), etc
        2. 'normal' for  normal field and 'skew' for skew field.
    """
    _magfuncs = {
        'dipole': {'type': 'normal', 'harmonic': 0},
        'corrector-horizontal': {'type': 'normal', 'harmonic': 0},
        'corrector-vertical': {'type': 'skew', 'harmonic': 0},
        'quadrupole': {'type': 'normal', 'harmonic': 1},
        'quadrupole-skew': {'type': 'skew', 'harmonic': 1},
        'sextupole': {'type': 'normal', 'harmonic': 2},
        'id-apu': {'type': 'normal', 'harmonic': 0},
        'lens': {'type': 'normal', 'harmonic': 1},
    }
    return _magfuncs


def get_multipole_name(harmonic, suffix=None):
    """Return names of multipoles, given harmonic number."""
    names = {0: 'dipole',
             1: 'quadrupole',
             2: 'sextupole',
             3: 'octupole',
             4: 'decapole',
             5: 'dodecapole', }
    suffix = 'pole' if suffix is None else suffix
    try:
        return names[harmonic].replace('pole', suffix)
    except KeyError:
        return '{0:d}-pole'.format((harmonic+1)*2).replace('pole', suffix)


def get_multipole_si_units(harmonic, power=None, product=None):
    """Return multipole units, given harmonic number."""
    product = '.' if product is None else product
    if harmonic == 0:
        return 'T{0:s}m'.format(product)
    elif harmonic == 1:
        return 'T'
    elif harmonic == 2:
        return 'T/m'
    else:
        power = '^' if power is None else power
        return 'T/m{0:s}{1:d}'.format(power, harmonic-1)


def linear_interpolation(xvals, xtab, ytab):
    """Return linear interpolation function value."""
    interp = _numpy.interp(
        xvals, xtab, ytab, left=-_numpy.inf, right=_numpy.inf)
    neg = _numpy.isneginf(interp)
    pos = _numpy.isposinf(interp)
    if neg.any():
        interp[neg] = linear_extrapolation(
            xvals[neg], xtab[0], xtab[1], ytab[0], ytab[1])
    if pos.any():
        interp[pos] = linear_extrapolation(
            xvals[pos], xtab[-1], xtab[-2], ytab[-1], ytab[-2])
    return interp


def linear_extrapolation(x, x1, x2, y1, y2):
    """Return linear extrapolation function value."""
    if x2 == x1:
        return min(y1, y2, key=abs)
    else:
        return y1 + (y2-y1)*(x-x1)/(x2-x1)


def sum_magnetic_multipoles(*multipoles_list):
    """Sum an iterable composed of multipoles dictionaries."""
    res = {}
    for multipoles in multipoles_list:
        nmpoles = multipoles.get('normal', {})
        res_tmp = res.get('normal', {})
        for harm, mpole in nmpoles.items():
            res_tmp[harm] = res_tmp.get(harm, 0.0) + mpole
        if res_tmp:
            res['normal'] = res_tmp
        smpoles = multipoles.get('skew', {})
        res_tmp = res.get('skew', {})
        for harm, mpole in smpoles.items():
            res_tmp[harm] = res_tmp.get(harm, 0.0) + mpole
        if res_tmp:
            res['skew'] = res_tmp
    return res


# TODO: merge this function with corresponding functions|class in
# ramp. also, default nrpts should be taken from csdevices!!!
def get_default_ramp_waveform(interval=500, nrpts=4000,
                              ti=None, fi=None, forms=None):
    """Generate normalized ramp."""
    time = interval * _numpy.linspace(0, 1.0, nrpts)
    if ti is None:
        ti = interval * _numpy.array([0, 13, 310,
                                      322, 330, 342, 480, 500])/500.0
    if fi is None:
        fi = _numpy.array([0.01, 0.02625, 1.0339285714,
                           1.05, 1.05, 1, 0.07, 0.01])
    if forms is None:
        forms = ['cube', 'injection', 'cube', 'line', 'cube', 'line', 'cube']

    # brho_3gev = 10.00692271077752  # [T.m]
    # brho_150mev = 0.5003432394479871  # [T.m]

    ramp = _numpy.zeros(len(time))
    va1 = _numpy.zeros(len(ti))
    dti = _numpy.zeros(len(forms))

    # findout the initial derivatives and delta time from the straigth lines
    for i, wfm in enumerate(forms):
        dti[i] = (ti[i+1]-ti[i])
        if wfm.startswith(('l', 'i')) or wfm == 0:
            equal_previous = False
            va1[i] = (fi[i+1]-fi[i])/dti[i]
            va1[i+1] = va1[i]
        else:
            equal_previous = True
        if equal_previous:
            va1[i] = va1[i-1]

    # calculate the ramp:
    for i, wfm in enumerate(forms):
        ind = _numpy.bitwise_and(ti[i] <= time, time <= ti[i+1])
        dtime = time[ind] - ti[i]
        va2, va3 = 0, 0
        if wfm.startswith(('c', 'i')):
            va2 = (-3*(fi[i]-fi[i+1]) - dti[i] * (2*va1[i]+va1[i+1]))/dti[i]**2
            va3 = (2*(fi[i]-fi[i+1]) + dti[i] * (va1[i]+va1[i+1]))/dti[i]**3
        ramp[ind] = fi[i] + va1[i]*dtime + va2*dtime**2 + va3*dtime**3

    return ramp


def get_section_dipole_name(maname):
    """Return name of dipole in the same section of given magnet name."""
    maname = _SiriusPVName(maname)
    if _re.match('B.*', maname.dev):
        return None
    elif maname.sec == 'SI':
        if maname.dev in {"InjDpKckr", "InjNLKckr"}:
            return 'TS-Fam:MA-B'
        return "SI-Fam:MA-B1B2"
    elif maname.sec == 'BO':
        if maname.dev in {'InjKckr', 'EjeKckr'}:
            return 'TB-Fam:MA-B'
        return 'BO-Fam:MA-B'
    elif maname.sec in {'TB', 'LI'}:
        return 'TB-Fam:MA-B'
    elif maname.sec == 'TS':
        return 'TS-Fam:MA-B'
    else:
        raise NotImplementedError(
            'No section named {}'.format(maname.sec))


def get_magnet_family_name(maname):
    """Return family name associated with a given magnet name."""
    maname = _SiriusPVName(maname)
    if maname.sec == "SI" and \
            maname.sub != "Fam" and \
            _re.match("(?:QD|QF|Q[0-9]).*", maname.dev):
        return _re.sub("SI-\d{2}\w{2}:", "SI-Fam:", maname)
    else:
        return None


def magnet_class(maname):
    """Return class of a magnet: Dipole, Normal, Trim, Pulsed."""
    maname = _SiriusPVName(maname)
    if maname.dis == 'ID':
        if maname.dev in ('APU22', 'APU58', ):
            return 'id-apu'
    if maname.dis != 'MA' and maname.dis != 'PM':
        raise ValueError("Cannot classify {}".format(maname))
    if maname.dis == 'PM':
        return 'pulsed'
    if maname.dev.startswith('B'):
        return 'dipole'
    if 'QS' not in maname.dev and \
            maname.sec == 'SI' and \
            maname.dev.startswith('Q') and \
            maname.sub != 'Fam':
        return 'trim'

    return 'normal'



"""Definition of thread classes to used across the package."""

import time as _time
from threading import Thread as _Thread, Event as _Event, \
    Lock as _Lock
from queue import Queue as _Queue, Empty as _Empty, Full as _Full

from epics.ca import use_initial_context as _use_initial_context

from .epics import CAThread as _CAThread


class AsyncWorker(_Thread):
    """Asynchronous Worker Thread.

    Performs asynchronous jobs indefinitely.
    """

    def __init__(self, name=None, is_cathread=False):
        """."""
        super().__init__(name=name, daemon=True)
        self.is_cathread = is_cathread
        self._evt_received = _Event()
        self._evt_ready = _Event()
        self._evt_ready.set()
        self._evt_stop = _Event()
        self.target = None
        self.args = tuple()

    def configure_new_run(self, target, args=None):
        """Configure a new run of the thread."""
        if not self._evt_ready.is_set():
            return False
        self.target = target
        self.args = args or tuple()
        self._evt_ready.clear()
        # NOTE: _evt_received setting must be last operation of this method.
        self._evt_received.set()
        return True

    def wait_ready(self, timeout=None):
        """Wait until last run is finished."""
        self._evt_ready.wait(timeout=timeout)

    def is_ready(self):
        """Check if last run has finished."""
        return self._evt_ready.is_set()

    def stop(self):
        """Stop thread."""
        self._evt_stop.set()

    def run(self):
        """."""
        if self.is_cathread:
            _use_initial_context()
        while not self._evt_stop.is_set():
            if self._evt_received.wait(0.5):
                self._evt_received.clear()
                self.target(*self.args)
                self._evt_ready.set()


class RepeaterThread(_Thread):
    """Repeat execution of predefined function for a given number of times."""

    def __init__(
            self, interval, function, args=None, kwargs=None, niter=0,
            is_cathread=False):
        """Init method.

        Inputs:
        - interval: time to wait between sucessive calls to method.
        - function: method to be executed.
        - args: tuple of positional arguments of method
        - kwargs: dictionary with keyword arguments of method.
        - niter: number of times to execute method. If niter is zero or None
            it will be executed indefinetly until stop is called.
        - whether to use CAThread logic .
        """
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = dict()
        super().__init__(daemon=True)
        self.is_cathread = is_cathread
        self.interval = interval
        if not hasattr(function, '__call__'):
            raise TypeError('Argument "function" is not callable.')
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.niters = niter
        self.cur_iter = 0
        self._stopped = _Event()
        self._unpaused = _Event()
        self._unpaused.set()

    def run(self):
        """Run method."""
        if self.is_cathread:
            _use_initial_context()

        self._unpaused.wait()
        self.function(*self.args, **self.kwargs)
        dtime = 0.0
        while ((not self._stopped.wait(self.interval - dtime)) and
               (not self.niters or self.niters > self.cur_iter)):
            self._unpaused.wait()
            if self._stopped.is_set():
                break
            self.cur_iter += 1
            _t0 = _time.time()
            self.function(*self.args, **self.kwargs)
            dtime = _time.time() - _t0

    def reset(self):
        """Reset count."""
        self.cur_iter = 0

    def pause(self):
        """."""
        self._unpaused.clear()

    def resume(self):
        """."""
        self._unpaused.set()

    def unpause(self):
        """."""
        self.resume()

    def is_paused(self):
        """."""
        return not self._unpaused.is_set()

    def isPaused(self):
        """."""
        return self.is_paused()

    def stop(self):
        """Stop execution."""
        self._unpaused.set()
        self._stopped.set()


class _BaseQueueThread(_Queue):
    """QueueThread.

    This class manages generic operations (actions) using a FIFO queue. Each
    operation processing is a method invoked as a separate thread. It also has
    process loop methods to continuously process its operation queue in a
    single thread.
    """

    def __init__(self, is_cathread=False):
        """Init."""
        super().__init__()
        self.is_cathread = is_cathread
        self._last_operation = None
        self._th = None
        self._changing_queue = _Lock()

    @property
    def last_operation(self):
        """Return last operation."""
        return self._last_operation

    @property
    def is_running(self):
        """."""
        return self._th is not None and self._th.is_alive()

    def put(self, operation, block=True, timeout=None):
        """Put operation to queue."""
        if not hasattr(operation, '__len__'):
            operation = (operation, )
        if not hasattr(operation[0], '__call__'):
            raise TypeError(
                'First element of "operation" is not callable.')

        with self._changing_queue:
            try:
                super().put(operation, block=block, timeout=timeout)
            except _Full:
                return False
            self._last_operation = operation
        return True

    def get(self, block=True, timeout=None):
        """Get operation from queue."""
        with self._changing_queue:
            try:
                return super().get(block=block, timeout=timeout)
            except _Empty:
                return None

    def _get_operation(self, block=True, timeout=None):
        """Raise queue.Empty exception in case of timeout or empty Queue."""
        operation = super().get(block=block, timeout=timeout)
        # process operation taken from queue
        args, kws = tuple(), dict()
        if len(operation) == 1:
            func = operation[0]
        elif len(operation) == 2:
            func, args = operation
        elif len(operation) >= 3:
            func, args, kws = operation[:3]
        return func, args, kws

    def _run_process(self, func, args, kws):
        """."""
        try:
            func(*args, **kws)
        except Exception:
            pass
        self.task_done()


class QueueThreads(_BaseQueueThread):
    """QueueThreads.

    This class manages generic operations (actions) using a FIFO queue. Each
    operation processing is a method invoked as a separate thread.
    """

    def process(self, block=True, timeout=None):
        """Process operation from queue."""
        with self._changing_queue:
            # first check if a thread is already running
            if self.is_running:
                return False
            # no thread is running, we can process queue
            try:
                func, args, kws = self._get_operation(block, timeout)
            except _Empty:
                return False
            th_cls = _CAThread if self.is_cathread else _Thread
            self._th = th_cls(
                target=self._run_process, args=(func, args, kws), daemon=True)
            self._th.start()
            return True


class LoopQueueThread(_BaseQueueThread):
    """LoopQueueThread.

    This class manages generic operations (actions) using a FIFO queue.
    It has a process loop to continuously process its operation queue in a
    single thread.
    """

    def __init__(self, is_cathread=False):
        """Init."""
        super().__init__(is_cathread=is_cathread)
        self._loop_stop_evt = _Event()

    def start(self):
        """Run deque process loop."""
        # check if there is a previously running loop
        self._loop_stop_evt.clear()
        if self.is_running:
            return
        # start new process loop
        th_cls = _CAThread if self.is_cathread else _Thread
        self._th = th_cls(target=self._loop_run_process, daemon=True)
        self._th.start()

    def stop(self):
        """Stop deque process loop."""
        self._loop_stop_evt.set()

    def _loop_run_process(self):
        while not self._loop_stop_evt.is_set():
            try:
                func, args, kws = self._get_operation(True, timeout=1)
                self._run_process(func, args, kws)
            except _Empty:
                continue
        self._loop_stop_evt.clear()



#!/usr/bin/env python-sirius

"""Unittest module for csdev.py."""

from unittest import TestCase
import siriuspy.util as util
import siriuspy.csdev as csdev

PUB_INTERFACE = (
    'ETypes',
    'Const',
    'add_pvslist_cte',
    'get_device_2_ioc_ip',
)


class TestUtil(TestCase):
    """Test util module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            csdev, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_adds_pvslist_cte(self):
        """Test adds_pvslist_cte."""
        dbase = {'a': {}, 'b': {}}
        dbase = csdev.add_pvslist_cte(dbase)
        self.assertEqual(len(dbase), 3)
        self.assertIn('Properties-Cte', dbase)
        self.assertEqual(dbase['Properties-Cte']['count'], 18)
        self.assertEqual(dbase['Properties-Cte']['value'], 'Properties-Cte a b')

    def test_get_device_2_ioc_ip(self):
        """Test get_device_2_ioc_ip."""
        # TODO: implement!



"""BSMP types definitions."""
import enum
import typing


@enum.unique
class BSMPBytesFmt(str, enum.Enum):
    CHAR = "<c"
    UCHAR = "<B"
    USHORT = "<H"
    UINT = "<I"
    FLOAT = "<f"


class BSMPType(typing.NamedTuple):
    type: str
    size: int
    fmt: str
    check: typing.Callable[[typing.Any], bool]


class Types:
    """Types used by BSMP entities."""

    T_CHAR = BSMPType("char", 1, BSMPBytesFmt.CHAR.value, lambda x: isinstance(x, str))
    T_UINT8 = BSMPType("uint8", 1, BSMPBytesFmt.UCHAR.value, lambda x: isinstance(x, int))
    T_UINT16 = BSMPType("uint16", 2, BSMPBytesFmt.USHORT.value, lambda x: isinstance(x, int))
    T_UINT32 = BSMPType("uint32", 4, BSMPBytesFmt.UINT.value, lambda x: isinstance(x, int))
    T_FLOAT = BSMPType("float", 4, BSMPBytesFmt.FLOAT.value, lambda x: isinstance(x, float))
    T_PARAM = BSMPType("uint16", 2, BSMPBytesFmt.USHORT.value, lambda x: isinstance(x, int))
    T_STATE = BSMPType("uint16", 2, BSMPBytesFmt.USHORT.value, lambda x: isinstance(x, int))
    T_ENUM = BSMPType("uint16", 2, BSMPBytesFmt.USHORT.value, lambda x: isinstance(x, int))
    T_DSP_CLASS = BSMPType("uint16", 2, BSMPBytesFmt.USHORT.value, lambda x: isinstance(x, int))



"""Module with Ramp exceptions."""


class RampError(ValueError):
    """General Ramp Error."""


class RampInvalidDipoleWfmParms(RampError):
    """Invalid Dipole waveform parameters."""


class RampInvalidNormConfig(RampError):
    """Invalid normalized configuration."""


class RampInvalidRFParms(RampError):
    """Invalid RF parameters."""



"""Clientconfig subpackage."""

from .configdb_client import ConfigDBClient, ConfigDBException
from .configdb_document import ConfigDBDocument
from .pvsconfig import PVsConfig

del configdb_client, configdb_document, pvsconfig



"""BPM devices."""

import time as _time
from threading import Event as _Flag
import numpy as _np

from .device import Device as _Device, Devices as _Devices
from ..diagbeam.bpm.csdev import Const as _csbpm
from ..search import BPMSearch as _BPMSearch
from ..namesys import SiriusPVName as _PVName


class BPM(_Device):
    """BPM Device."""

    ACQSTATES_NOTOK = {
        _csbpm.AcqStates.Error, _csbpm.AcqStates.No_Memory,
        _csbpm.AcqStates.Too_Few_Samples,
        _csbpm.AcqStates.Too_Many_Samples, _csbpm.AcqStates.Acq_Overflow}
    ACQSTATES_STARTED = {
        _csbpm.AcqStates.Waiting, _csbpm.AcqStates.External_Trig,
        _csbpm.AcqStates.Data_Trig, _csbpm.AcqStates.Software_Trig,
        _csbpm.AcqStates.Acquiring}
    ACQSTATES_FINISHED = {_csbpm.AcqStates.Idle, _csbpm.AcqStates.Aborted}
    ACQSTATES_FINISHED |= ACQSTATES_NOTOK

    _properties = (
        'asyn.ENBL', 'asyn.CNCT', 'SwMode-Sel', 'SwMode-Sts',
        'RFFEAtt-SP', 'RFFEAtt-RB',
        'SP_AArrayData', 'SP_BArrayData', 'SP_CArrayData', 'SP_DArrayData',
        'GEN_AArrayData', 'GEN_BArrayData', 'GEN_CArrayData', 'GEN_DArrayData',
        'GEN_XArrayData', 'GEN_YArrayData', 'GEN_SUMArrayData',
        'GEN_QArrayData',
        'GEN_RawXArrayData', 'GEN_RawYArrayData', 'GEN_RawSUMArrayData',
        'GEN_RawQArrayData',
        'SPPosX-Mon', 'SPPosY-Mon', 'SPSum-Mon', 'SPPosQ-Mon',
        'SPAmplA-Mon', 'SPAmplB-Mon', 'SPAmplC-Mon', 'SPAmplD-Mon',
        'PosX-Mon', 'PosY-Mon', 'Sum-Mon', 'PosQ-Mon',
        'AmplA-Mon', 'AmplB-Mon', 'AmplC-Mon', 'AmplD-Mon',
        'INFOClkFreq-RB', 'INFOHarmonicNumber-RB', 'INFOTbTRate-RB',
        'INFOFOFBRate-RB', 'INFOMONITRate-RB', 'INFOFAcqRate-RB',
        'GEN_PolyXArrayCoeff-SP', 'GEN_PolyXArrayCoeff-RB',
        'GEN_PolyYArrayCoeff-SP', 'GEN_PolyYArrayCoeff-RB',
        'GEN_PolySUMArrayCoeff-SP', 'GEN_PolySUMArrayCoeff-RB',
        'GEN_PolyQArrayCoeff-SP', 'GEN_PolyQArrayCoeff-RB',
        'PosKx-SP', 'PosKx-RB',
        'PosKy-RB', 'PosKy-SP',
        'PosKsum-SP', 'PosKsum-RB',
        'PosKq-SP', 'PosKq-RB',
        'PosXOffset-SP', 'PosXOffset-RB',
        'PosYOffset-SP', 'PosYOffset-RB',
        'PosSumOffset-SP', 'PosSumOffset-RB',
        'PosQOffset-SP', 'PosQOffset-RB',
        'ACQBPMMode-Sel', 'ACQBPMMode-Sts',
        'ACQChannel-Sel', 'ACQChannel-Sts',
        'ACQShots-SP', 'ACQShots-RB',
        'ACQUpdateTime-SP', 'ACQUpdateTime-RB',
        'ACQSamplesPre-SP', 'ACQSamplesPre-RB',
        'ACQSamplesPost-SP', 'ACQSamplesPost-RB',
        'ACQTriggerEvent-Sel', 'ACQTriggerEvent-Sts',
        'ACQStatus-Sts', 'ACQCount-Mon',
        'ACQTrigger-Sel', 'ACQTrigger-Sts',
        'ACQTriggerRep-Sel', 'ACQTriggerRep-Sts',
        'ACQDataTrigChan-Sel', 'ACQDataTrigChan-Sts',
        'ACQTriggerDataSel-SP', 'ACQTriggerDataSel-RB',
        'ACQTriggerDataThres-SP', 'ACQTriggerDataThres-RB',
        'ACQTriggerDataPol-Sel', 'ACQTriggerDataPol-Sts',
        'ACQTriggerDataHyst-SP', 'ACQTriggerDataHyst-RB',
        'SwTagEn-Sel', 'SwTagEn-Sts', 'SwDivClk-RB',
        'TbTTagEn-Sel', 'TbTTagEn-Sts',
        'FAcqTagEn-Sel', 'FAcqTagEn-Sts',
        'MonitTagEn-Sel', 'MonitTagEn-Sts',
        'TbTDataMaskEn-Sel', 'TbTDataMaskEn-Sts',
        'TbTDataMaskSamplesBeg-SP', 'TbTDataMaskSamplesBeg-RB',
        'TbTDataMaskSamplesEnd-SP', 'TbTDataMaskSamplesEnd-RB',
        'XYPosCal-Sel', 'XYPosCal-Sts',
        'SUMPosCal-Sel', 'SUMPosCal-Sts',
        'QPosCal-Sel', 'QPosCal-Sts',
        )

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    def __init__(self, devname, auto_monitor_mon=True, ispost_mortem=False):
        """."""
        # call base class constructor
        self._ispost_mortem = ispost_mortem
        properties = {self.get_propname(p) for p in BPM._properties}

        if _BPMSearch.is_photon_bpm(devname):
            properties -= {'RFFEAtt-SP', 'RFFEAtt-RB'}
        properties = list(properties)

        super().__init__(
            devname, properties=properties, auto_monitor_mon=auto_monitor_mon)
        self.csdata = _csbpm

    def __str__(self):
        """."""
        stg = '################### Summary Status ###################\n'
        stg += 'asyn:\n'
        stg += f'    Enabled: {_csbpm.EnblTyp._fields[self.asyn_state]:s}\n'
        stg += '    Connected: '
        stg += '    Switching Mode: '
        stg += f'{_csbpm.SwModes._fields[self.switching_mode]:s}\n'
        stg += f'{_csbpm.ConnTyp._fields[self.asyn_connected]:s}\n'
        stg += '\nAcquisition Parameters:\n'
        stg += f'    - Status: {_csbpm.AcqStates._fields[self.acq_status]:s}\n'
        stg += f'    - Mode: {_csbpm.OpModes._fields[self.acq_mode]:s}\n'
        stg += f'    - Channel: {_csbpm.AcqChan._fields[self.acq_channel]:s}\n'
        stg += f'    - Nr Shots: {self.acq_nrshots:d}\n'
        stg += f'    - Update Time: {self.acq_update_time:.1f} ms\n'
        stg += f'    - Repeat: {_csbpm.AcqRepeat._fields[self.acq_repeat]:s}\n'
        stg += '    - Trigger Type: '
        stg += f'{_csbpm.AcqTrigTyp._fields[self.acq_trigger]:s}\n'
        if self.acq_trigger == _csbpm.AcqTrigTyp.Data:
            stg += '        - Channel: '
            stg += f'{_csbpm.AcqChan._fields[self.acq_trig_datachan]:s}\n'
            stg += '        - Source: '
            stg += f'{_csbpm.AcqDataTyp._fields[self.acq_trig_datasel]:s}\n'
            stg += '        - Polarity: '
            stg += f'{_csbpm.Polarity._fields[self.acq_trig_datapol]:s}\n'
            stg += f'        - Threshold: {self.acq_trig_datathres:.1f}\n'
            stg += f'        - Hysteresis: {self.acq_trig_datahyst:d}\n'
        stg += '\n'
        return stg

    @property
    def is_ok(self):
        """."""
        okay = self.acq_status not in self.ACQSTATES_NOTOK
        okay &= self.asyn_connected == _csbpm.ConnTyp.Connected
        okay &= self.asyn_state == _csbpm.EnblTyp.Enable
        return okay

    @property
    def rffe_att(self):
        """."""
        if 'RFFEAtt-RB' not in self._pvs:
            return None
        return self['RFFEAtt-RB']

    @rffe_att.setter
    def rffe_att(self, val):
        """."""
        if 'RFFEAtt-SP' in self._pvs:
            self['RFFEAtt-SP'] = val

    @property
    def asyn_state(self):
        """."""
        return self['asyn.ENBL']

    @asyn_state.setter
    def asyn_state(self, boo):
        """."""
        val = _csbpm.EnblTyp.Enable if boo else _csbpm.EnblTyp.Disable
        self['asyn.ENBL'] = val

    @property
    def asyn_connected(self):
        """."""
        return self['asyn.CNCT']

    @property
    def switching_mode(self):
        """."""
        return self['SwMode-Sts']

    @switching_mode.setter
    def switching_mode(self, val):
        """."""
        self['SwMode-Sel'] = val

    @property
    def switching_mode_str(self):
        """."""
        return _csbpm.SwModes._fields[self.switching_mode]

    @property
    def switching_rate(self):
        """Switching rate manifested at BPM readings.

        NOTE: Actually this property maps the rate of the effect of
        switching on the orbit, which is twice the switching rate,
        because the switching has two states.

        """
        return self['SwDivClk-RB'] * 2

    @property
    def switching_period(self):
        """Switching period manifested at BPM readings."""
        return self.switching_rate / self.adcfreq

    @property
    def harmonic_number(self):
        """."""
        return self['INFOHarmonicNumber-RB']

    @property
    def adcfreq(self):
        """."""
        return self['INFOClkFreq-RB']

    @property
    def tbt_rate(self):
        """Divisor or TbT in relation to ADC."""
        return self['INFOTbTRate-RB']

    @property
    def tbt_period(self):
        """."""
        return self.tbt_rate / self.adcfreq

    @property
    def tbt_sync_enbl(self):
        """."""
        return self['TbTTagEn-Sts']

    @tbt_sync_enbl.setter
    def tbt_sync_enbl(self, val):
        """."""
        self['TbTTagEn-Sel'] = val

    @property
    def tbt_mask_enbl(self):
        """."""
        return self['TbTDataMaskEn-Sts']

    @tbt_mask_enbl.setter
    def tbt_mask_enbl(self, val):
        """."""
        self['TbTDataMaskEn-Sel'] = val

    @property
    def tbt_mask_beg(self):
        """."""
        return self['TbTDataMaskSamplesBeg-RB']

    @tbt_mask_beg.setter
    def tbt_mask_beg(self, val):
        """."""
        self['TbTDataMaskSamplesBeg-SP'] = val

    @property
    def tbt_mask_end(self):
        """."""
        return self['TbTDataMaskSamplesEnd-RB']

    @tbt_mask_end.setter
    def tbt_mask_end(self, val):
        """."""
        self['TbTDataMaskSamplesEnd-SP'] = val

    @property
    def fofb_sync_enbl(self):
        """."""
        return self['SwTagEn-Sts']

    @fofb_sync_enbl.setter
    def fofb_sync_enbl(self, val):
        """."""
        self['SwTagEn-Sel'] = val

    @property
    def fofb_rate(self):
        """Divisor or FOFB in relation to ADC."""
        return self['INFOFOFBRate-RB']

    @property
    def fofb_period(self):
        """."""
        return self.fofb_rate / self.adcfreq

    @property
    def facq_rate(self):
        """Divisor or FAcq in relation to ADC."""
        return self['INFOFAcqRate-RB']

    @property
    def facq_period(self):
        """."""
        return self.facq_rate / self.adcfreq

    @property
    def facq_sync_enbl(self):
        """."""
        return self['FAcqTagEn']

    @facq_sync_enbl.setter
    def facq_sync_enbl(self, val):
        """."""
        self['FAcqTagEn-Sel'] = val

    @property
    def monit_rate(self):
        """Divisor or Monit in relation to ADC."""
        return self['INFOMONITRate-RB']

    @property
    def monit_period(self):
        """."""
        return self.monit_rate / self.adcfreq

    @property
    def monit_sync_enbl(self):
        """."""
        return self['MonitTagEn-Sts']

    @monit_sync_enbl.setter
    def monit_sync_enbl(self, val):
        """."""
        self['MonitTagEn-Sel'] = val

    @property
    def posx_gain(self):
        """."""
        return self['PosKx-RB'] * self.CONV_NM2UM

    @posx_gain.setter
    def posx_gain(self, value):
        self['PosKx-SP'] = float(value) / self.CONV_NM2UM

    @property
    def posy_gain(self):
        """."""
        return self['PosKy-RB'] * self.CONV_NM2UM

    @posy_gain.setter
    def posy_gain(self, value):
        self['PosKy-SP'] = float(value) / self.CONV_NM2UM

    @property
    def possum_gain(self):
        """."""
        return self['PosKsum-RB']

    @possum_gain.setter
    def possum_gain(self, value):
        self['PosKsum-SP'] = float(value)

    @property
    def posq_gain(self):
        """."""
        return self['PosKq-RB']

    @posq_gain.setter
    def posq_gain(self, value):
        self['PosKq-SP'] = float(value)

    @property
    def posx_offset(self):
        """."""
        return self['PosXOffset-RB'] * self.CONV_NM2UM

    @posx_offset.setter
    def posx_offset(self, value):
        self['PosXOffset-SP'] = float(value) / self.CONV_NM2UM

    @property
    def posy_offset(self):
        """."""
        return self['PosYOffset-RB'] * self.CONV_NM2UM

    @posy_offset.setter
    def posy_offset(self, value):
        self['PosYOffset-SP'] = float(value) / self.CONV_NM2UM

    @property
    def possum_offset(self):
        """."""
        return self['PosSumOffset-RB']

    @possum_offset.setter
    def possum_offset(self, value):
        self['PosSumOffset-SP'] = float(value)

    @property
    def posq_offset(self):
        """."""
        return self['PosQOffset-RB']

    @posq_offset.setter
    def posq_offset(self, value):
        self['PosQOffset-SP'] = float(value)

    @property
    def posx(self):
        """."""
        return self['PosX-Mon'] * self.CONV_NM2UM

    @property
    def posy(self):
        """."""
        return self['PosY-Mon'] * self.CONV_NM2UM

    @property
    def possum(self):
        """."""
        return self['Sum-Mon']

    @property
    def posq(self):
        """."""
        return self['PosQ-Mon']

    @property
    def ampla(self):
        """."""
        return self['AmplA-Mon']

    @property
    def amplb(self):
        """."""
        return self['AmplB-Mon']

    @property
    def amplc(self):
        """."""
        return self['AmplC-Mon']

    @property
    def ampld(self):
        """."""
        return self['AmplD-Mon']

    @property
    def mt_posx(self):
        """."""
        return self[self.get_propname('GEN_XArrayData')] * self.CONV_NM2UM

    @property
    def mt_posy(self):
        """."""
        return self[self.get_propname('GEN_YArrayData')] * self.CONV_NM2UM

    @property
    def mt_possum(self):
        """."""
        return self[self.get_propname('GEN_SUMArrayData')]

    @property
    def mt_posq(self):
        """."""
        return self[self.get_propname('GEN_QArrayData')]

    @property
    def mt_ampla(self):
        """."""
        return self[self.get_propname('GEN_AArrayData')]

    @property
    def mt_amplb(self):
        """."""
        return self[self.get_propname('GEN_BArrayData')]

    @property
    def mt_amplc(self):
        """."""
        return self[self.get_propname('GEN_CArrayData')]

    @property
    def mt_ampld(self):
        """."""
        return self[self.get_propname('GEN_DArrayData')]

    @property
    def mt_polyx(self):
        """."""
        return self[self.get_propname('GEN_PolyXArrayCoeff-RB')]

    @mt_polyx.setter
    def mt_polyx(self, value):
        """."""
        self[self.get_propname('GEN_PolyXArrayCoeff-SP')] = _np.array(value)

    @property
    def mt_polyy(self):
        """."""
        return self[self.get_propname('GEN_PolyYArrayCoeff-RB')]

    @mt_polyy.setter
    def mt_polyy(self, value):
        """."""
        self[self.get_propname('GEN_PolyYArrayCoeff-SP')] = _np.array(value)

    @property
    def mt_polysum(self):
        """."""
        return self[self.get_propname('GEN_PolySUMArrayCoeff-RB')]

    @mt_polysum.setter
    def mt_polysum(self, value):
        """."""
        self[self.get_propname('GEN_PolySUMArrayCoeff-SP')] = _np.array(value)

    @property
    def mt_polyq(self):
        """."""
        return self[self.get_propname('GEN_PolyQArrayCoeff-RB')]

    @mt_polyq.setter
    def mt_polyq(self, value):
        """."""
        self[self.get_propname('GEN_PolyQArrayCoeff-SP')] = _np.array(value)

    @property
    def mt_polyxy_enbl(self):
        """."""
        return self['XYPosCal-Sts']

    @mt_polyxy_enbl.setter
    def mt_polyxy_enbl(self, val):
        """."""
        self['XYPosCal-Sel'] = val

    @property
    def mt_polysum_enbl(self):
        """."""
        return self['SUMPosCal-Sts']

    @mt_polysum_enbl.setter
    def mt_polysum_enbl(self, val):
        """."""
        self['SUMPosCal-Sel'] = val

    @property
    def mt_polyq_enbl(self):
        """."""
        return self['QPosCal-Sts']

    @mt_polyq_enbl.setter
    def mt_polyq_enbl(self, val):
        """."""
        self['QPosCal-Sel'] = val

    @property
    def mtraw_posx(self):
        """Multi turn raw X array data."""
        return self[self.get_propname('GEN_RawXArrayData')] * self.CONV_NM2UM

    @property
    def mtraw_posy(self):
        """Multi turn raw Y array data."""
        return self[self.get_propname('GEN_RawYArrayData')] * self.CONV_NM2UM

    @property
    def mtraw_possum(self):
        """Multi turn raw sum array data."""
        return self[self.get_propname('GEN_RawSUMArrayData')]

    @property
    def mtraw_posq(self):
        """Multi turn raw Q array data."""
        return self[self.get_propname('GEN_RawQArrayData')]

    @property
    def sp_posx(self):
        """."""
        return self['SPPosX-Mon'] * self.CONV_NM2UM

    @property
    def sp_posy(self):
        """."""
        return self['SPPosY-Mon'] * self.CONV_NM2UM

    @property
    def sp_possum(self):
        """."""
        return self['SPSum-Mon']

    @property
    def sp_posq(self):
        """."""
        return self['SPPosQ-Mon']

    @property
    def sp_ampla(self):
        """."""
        return self['SPAmplA-Mon']

    @property
    def sp_amplb(self):
        """."""
        return self['SPAmplB-Mon']

    @property
    def sp_amplc(self):
        """."""
        return self['SPAmplC-Mon']

    @property
    def sp_ampld(self):
        """."""
        return self['SPAmplD-Mon']

    @property
    def sp_arraya(self):
        """."""
        return self['SP_AArrayData']

    @property
    def sp_arrayb(self):
        """."""
        return self['SP_BArrayData']

    @property
    def sp_arrayc(self):
        """."""
        return self['SP_CArrayData']

    @property
    def sp_arrayd(self):
        """."""
        return self['SP_DArrayData']

    @property
    def acq_mode(self):
        """."""
        return self[self.get_propname('ACQBPMMode-Sts')]

    @acq_mode.setter
    def acq_mode(self, mode):
        """."""
        self[self.get_propname('ACQBPMMode-Sel')] = mode

    @property
    def acq_ctrl(self):
        """."""
        return self[self.get_propname('ACQTriggerEvent-Sts')]

    @acq_ctrl.setter
    def acq_ctrl(self, val):
        """."""
        self[self.get_propname('ACQTriggerEvent-Sel')] = val

    @property
    def acq_status(self):
        """."""
        return self[self.get_propname('ACQStatus-Sts')]

    @property
    def acq_count(self):
        """Counter of number of acquisitions so far."""
        return self[self.get_propname('ACQCount-Mon')]

    @property
    def acq_channel(self):
        """."""
        return self[self.get_propname('ACQChannel-Sts')]

    @acq_channel.setter
    def acq_channel(self, val):
        """."""
        self[self.get_propname('ACQChannel-Sel')] = val

    @property
    def acq_channel_str(self):
        """."""
        return _csbpm.AcqChan._fields[self.acq_channel]

    @property
    def acq_trigger(self):
        """."""
        return self[self.get_propname('ACQTrigger-Sts')]

    @acq_trigger.setter
    def acq_trigger(self, val):
        """."""
        self[self.get_propname('ACQTrigger-Sel')] = val

    @property
    def acq_repeat(self):
        """."""
        return self[self.get_propname('ACQTriggerRep-Sts')]

    @acq_repeat.setter
    def acq_repeat(self, val):
        """."""
        self[self.get_propname('ACQTriggerRep-Sel')] = val

    @property
    def acq_update_time(self):
        """."""
        return self[self.get_propname('ACQUpdateTime-RB')]  / 1e3

    @acq_update_time.setter
    def acq_update_time(self, val):
        """."""
        self[self.get_propname('ACQUpdateTime-SP')] = val * 1e3

    @property
    def acq_trig_datachan(self):
        """."""
        return self[self.get_propname('ACQDataTrigChan-Sts')]

    @acq_trig_datachan.setter
    def acq_trig_datachan(self, val):
        """."""
        self[self.get_propname('ACQDataTrigChan-Sel')] = val

    @property
    def acq_trig_datasel(self):
        """."""
        return self[self.get_propname('ACQTriggerDataSel-RB')]

    @acq_trig_datasel.setter
    def acq_trig_datasel(self, val):
        """."""
        self[self.get_propname('ACQTriggerDataSel-SP')] = val

    @property
    def acq_trig_datathres(self):
        """."""
        return self[self.get_propname('ACQTriggerDataThres-RB')]

    @acq_trig_datathres.setter
    def acq_trig_datathres(self, val):
        """."""
        self[self.get_propname('ACQTriggerDataThres-SP')] = val

    @property
    def acq_trig_datahyst(self):
        """."""
        return self[self.get_propname('ACQTriggerDataHyst-RB')]

    @acq_trig_datahyst.setter
    def acq_trig_datahyst(self, val):
        """."""
        self[self.get_propname('ACQTriggerDataHyst-SP')] = val

    @property
    def acq_trig_datapol(self):
        """."""
        return self[self.get_propname('ACQTriggerDataPol-RB')]

    @acq_trig_datapol.setter
    def acq_trig_datapol(self, val):
        """."""
        self[self.get_propname('ACQTriggerDataPol-SP')] = val

    @property
    def acq_nrsamples_post(self):
        """."""
        return self[self.get_propname('ACQSamplesPost-RB')]

    @acq_nrsamples_post.setter
    def acq_nrsamples_post(self, val):
        """."""
        self[self.get_propname('ACQSamplesPost-SP')] = val

    @property
    def acq_nrsamples_pre(self):
        """."""
        return self[self.get_propname('ACQSamplesPre-RB')]

    @acq_nrsamples_pre.setter
    def acq_nrsamples_pre(self, val):
        """."""
        self[self.get_propname('ACQSamplesPre-SP')] = val

    @property
    def acq_nrshots(self):
        """."""
        return self[self.get_propname('ACQShots-RB')]

    @acq_nrshots.setter
    def acq_nrshots(self, val):
        """."""
        self[self.get_propname('ACQShots-SP')] = val

    def wait_acq_finish(self, timeout=10):
        """Wait Acquisition to finish."""
        return self._wait(
            self.get_propname('ACQStatus-Sts'), self.ACQSTATES_FINISHED,
            timeout=timeout, comp=lambda x, y: x in y)

    def wait_acq_start(self, timeout=10):
        """Wait Acquisition to start."""
        return self._wait(
            self.get_propname('ACQStatus-Sts'), self.ACQSTATES_STARTED,
            timeout=timeout, comp=lambda x, y: x in y)

    def cmd_acq_start(self):
        """Command Start Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Start
        return self._wait(
            self.get_propname('ACQTriggerEvent-Sts'), _csbpm.AcqEvents.Start)

    def cmd_acq_stop(self):
        """Command Stop Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Stop
        return self._wait(
            self.get_propname('ACQTriggerEvent-Sts'), _csbpm.AcqEvents.Stop)

    def cmd_acq_abort(self):
        """Command Abort Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Abort
        return self._wait(
            self.get_propname('ACQTriggerEvent-Sts'), _csbpm.AcqEvents.Abort)

    def cmd_turn_on_switching(self):
        """Command Turn on Switching."""
        self.switching_mode = _csbpm.SwModes.switching
        return self._wait('SwMode-Sts', _csbpm.SwModes.switching)

    def cmd_turn_off_switching(self):
        """Command Turn off Switching."""
        self.switching_mode = _csbpm.SwModes.direct
        return self._wait('SwMode-Sts', _csbpm.SwModes.direct)

    def cmd_sync_tbt(self):
        """Synchronize TbT acquisitions with Timing System."""
        self.tbt_sync_enbl = 1
        _time.sleep(0.1)
        self.tbt_sync_enbl = 0
        return self._wait('TbTTagEn-Sts', 0)

    def cmd_sync_fofb(self):
        """Synchronize FOFB acquisitions with Timing System."""
        self.fofb_sync_enbl = 1
        _time.sleep(0.1)
        self.fofb_sync_enbl = 0
        return self._wait('SwTagEn-Sts', 0)

    def cmd_sync_facq(self):
        """Synchronize FAcq acquisitions with Timing System."""
        self.facq_sync_enbl = 1
        _time.sleep(0.1)
        self.facq_sync_enbl = 0
        return self._wait('FAcqTagEn-Sts', 0)

    def cmd_sync_monit(self):
        """Synchronize Monit acquisitions with Timing System."""
        self.monit_sync_enbl = 1
        _time.sleep(0.1)
        self.monit_sync_enbl = 0
        return self._wait('FAcqTagEn-Sts', 0)

    def get_sampling_frequency(
            self, rf_freq: float, acq_rate='') -> float:
        """Return the sampling frequency of the acquisition.

        Args:
            rf_freq (float): RF frequency.
            acq_rate (str, optional): acquisition rate. Defaults to ''.
            If empty string, it gets the configured acq. rate on BPMs

        Returns:
            float: acquisition frequency.

        """
        acq_rate = self.acq_channel_str if not acq_rate else acq_rate
        fadc = rf_freq / self.harmonic_number * self.tbt_rate
        if acq_rate.lower().startswith('tbt'):
            return fadc / self.tbt_rate
        elif acq_rate.lower().startswith('fofb'):
            return fadc / self.fofb_rate
        elif acq_rate.lower().startswith('facq'):
            return fadc / self.facq_rate
        return fadc / self.monit_rate

    def get_switching_frequency(self, rf_freq: float) -> float:
        """Return the switching frequency.

        Args:
            rf_freq (float): RF frequency.

        Returns:
            float: switching frequency.

        """
        fadc = rf_freq / self.harmonic_number * self.tbt_rate
        return fadc / self.switching_rate

    def get_propname(self, prop):
        if not self._ispost_mortem:
            return prop
        if prop.startswith('GEN'):
            return prop.replace('GEN', 'PM')
        elif prop.startswith('ACQ'):
            return prop.replace('ACQ', 'ACQ_PM')
        return prop


class FamBPMs(_Devices):
    """Family of BPMs.

    Parameters
    ----------
        devname (str, optional)
            Device name. If not provided, defaults to DEVICES.SI.
            Determine the list of BPM names.
        bpmnames ((list, tuple), optional)
            BPM names list. If provided, it takes priority over 'devname'
            parameter. Defaults to None.
        ispost_mortem (bool, optional)
            Whether to control PM acquisition core. Defaults to False.
    """

    TIMEOUT = 10
    RFFEATT_MAX = 30

    class DEVICES:
        """."""

        SI = 'SI-Fam:DI-BPM'
        BO = 'BO-Fam:DI-BPM'
        ALL = (BO, SI)

    def __init__(self, devname=None, bpmnames=None, ispost_mortem=False):
        """."""
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')

        devname = _PVName(devname)
        bpm_names = bpmnames or _BPMSearch.get_names(
            filters={'sec': devname.sec, 'dev': devname.dev})
        self._ispost_mortem = ispost_mortem
        devs = [
            BPM(dev, auto_monitor_mon=False, ispost_mortem=ispost_mortem)
            for dev in bpm_names]

        super().__init__(devname, devs)
        self._bpm_names = bpm_names
        self._csbpm = devs[0].csdata
        self._initial_timestamps = None

        self._mturn_flags = dict()
        # NOTE: ACQCount-Mon need to be fixed on BPM's IOC
        # for bpm in devs:
        #     pvo = bpm.pv_object(bpm.get_propname('ACQCount-Mon'))
        #     pvo.auto_monitor = True
        #     self._mturn_flags[pvo.pvname] = _Flag()
        #     pvo.add_callback(self._mturn_set_flag)

    @property
    def bpm_names(self):
        """Return BPM names."""
        return self._bpm_names

    @property
    def csbpm(self):
        """Return control system BPM constants class."""
        return self._csbpm

    def set_attenuation(self, value=RFFEATT_MAX, timeout=TIMEOUT):
        """."""
        for bpm in self:
            bpm.rffe_att = value

        mstr = ''
        okall = True
        t0 = _time.time()
        for bpm in self:
            tout = timeout - (_time.time() - t0)
            if not bpm._wait('RFFEAtt-RB', value, timeout=tout):
                okall = False
                mstr += (
                    f'\n{bpm.devname:<20s}: ' +
                    f'rb {bpm.rffe_att:.0f} != sp {value:.0f}')

        print('RFFE attenuation set confirmed in all BPMs', end='')
        print(', except:' + mstr if mstr else '.')
        return okall

    def get_slow_orbit(self):
        """Get slow orbit vectors.

        Returns:
            orbx (numpy.ndarray, 160): Horizontal Orbit.
            orby (numpy.ndarray, 160): Vertical Orbit.

        """
        orbx, orby = [], []
        for bpm in self._devices:
            orbx.append(bpm.posx)
            orby.append(bpm.posy)
        orbx = _np.array(orbx)
        orby = _np.array(orby)
        return orbx, orby

    def get_mturn_orbit(self, return_sum=False):
        """Get Multiturn orbit matrices.

        Args:
            return_sum (bool, optional): Whether or not to return BPMs sum.
                Defaults to False.

        Returns:
            orbx (numpy.ndarray, Nx160): Horizontal Orbit.
            orby (numpy.ndarray, Nx160): Vertical Orbit.
            possum (numpy.ndarray, Nx160): BPMs Sum signal.

        """
        orbx, orby = [], []
        if return_sum:
            possum = []

        mini = None
        for bpm in self._devices:
            mtx = bpm.mt_posx
            mty = bpm.mt_posy
            orbx.append(mtx)
            orby.append(mty)

            if mini is None:
                mini = mtx.size
            mini = _np.min([mini, mtx.size, mty.size])

            if return_sum:
                mts = bpm.mt_possum
                possum.append(mts)
                mini = min(mini, mts.size)

        for i, (obx, oby) in enumerate(zip(orbx, orby)):
            orbx[i] = obx[:mini]
            orby[i] = oby[:mini]
            if return_sum:
                possum[i] = possum[i][:mini]
        orbx = _np.array(orbx).T
        orby = _np.array(orby).T

        if not return_sum:
            return orbx, orby
        return orbx, orby, _np.array(possum).T

    def get_mturn_timestamps(self, return_sum=False):
        """Get Multiturn data timestamps.

        Args:
            return_sum (bool, optional): Whether or not to return BPMs sum
                timestamps. Defaults to False.

        Returns:
            tsmps (numpy.ndarray, (160, N)): The i-th row has the timestamp of
                the i-th bpm for the [horizontal, vertical, sum] signals
                respectively. If return_sum is False, then N=2 instead of 3.

        """
        tsmps = _np.zeros((len(self._devices), 2+return_sum), dtype=float)
        for i, bpm in enumerate(self._devices):
            pvx = bpm.pv_object(bpm.get_propname('GEN_XArrayData'))
            pvy = bpm.pv_object(bpm.get_propname('GEN_YArrayData'))
            vax = pvx.get_timevars(timeout=self.TIMEOUT)
            vay = pvy.get_timevars()
            tsmps[i, 0] = pvx.timestamp if vax is None else vax['timestamp']
            tsmps[i, 1] = pvy.timestamp if vay is None else vay['timestamp']
            if not return_sum:
                continue
            pvs = bpm.pv_object(bpm.get_propname('GEN_SUMArrayData'))
            vas = pvs.get_timevars()
            tsmps[i, 2] = pvs.timestamp if vas is None else vas['timestamp']
        return tsmps

    def get_sampling_frequency(self, rf_freq: float, acq_rate='') -> float:
        """Return the sampling frequency of the acquisition.

        Args:
            rf_freq (float): RF frequency.
            acq_rate (str, optional): acquisition rate. Defaults to ''.
            If empty string, it gets the configured acq. rate on BPMs

        Returns:
            float: acquisition frequency.

        """
        fs_bpms = {
            dev.get_sampling_frequency(rf_freq, acq_rate)
            for dev in self.devices}
        if len(fs_bpms) == 1:
            return fs_bpms.pop()
        else:
            print('BPMs are not configured with the same ACQChannel.')
            return None

    def get_switching_frequency(self, rf_freq: float) -> float:
        """Return the switching frequency.

        Args:
            rf_freq (float): RF frequency.

        Returns:
            float: switching frequency.

        """
        fsw_bpms = {
            dev.get_switching_frequency(rf_freq) for dev in self.devices}
        if len(fsw_bpms) == 1:
            return fsw_bpms.pop()
        else:
            print('BPMs are not configured with the same SwMode.')
            return None

    def mturn_config_acquisition(
            self, nr_points_after: int, nr_points_before=0,
            acq_rate='FAcq', repeat=True, external=True) -> int:
        """Configure acquisition for BPMs.

        Args:
            nr_points_after (int): number of points after trigger.
            nr_points_before (int): number of points after trigger.
                Defaults to 0.
            acq_rate (str, optional): Acquisition rate ('TbT', 'TbTPha',
                'FOFB', 'FOFBPha', 'FAcq', 'ADC', 'ADCSwp').
                Defaults to 'FAcq'.
            repeat (bool, optional): Whether or not acquisition should be
                repetitive. Defaults to True.
            external (bool, optional): Whether or not external trigger should
                be used. Defaults to True.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                <0: Index of the first BPM which did not stop last acq. plus 1.
                >0: Index of the first BPM which is not ready for acq. plus 1.

        """
        if acq_rate.lower().startswith('facq'):
            acq_rate = self._csbpm.AcqChan.FAcq
        elif acq_rate.lower().startswith('fofbpha'):
            acq_rate = self._csbpm.AcqChan.FOFBPha
        elif acq_rate.lower().startswith('fofb'):
            acq_rate = self._csbpm.AcqChan.FOFB
        elif acq_rate.lower().startswith('tbtpha'):
            acq_rate = self._csbpm.AcqChan.TbTPha
        elif acq_rate.lower().startswith('tbt'):
            acq_rate = self._csbpm.AcqChan.TbT
        elif acq_rate.lower().startswith('adcswp'):
            acq_rate = self._csbpm.AcqChan.ADCSwp
        elif acq_rate.lower().startswith('adc'):
            acq_rate = self._csbpm.AcqChan.ADC
        else:
            raise ValueError(acq_rate + ' is not a valid acquisition rate.')

        if repeat:
            repeat = self._csbpm.AcqRepeat.Repetitive
        else:
            repeat = self._csbpm.AcqRepeat.Normal

        if external:
            trig = self._csbpm.AcqTrigTyp.External
        else:
            trig = self._csbpm.AcqTrigTyp.Now

        ret = self.cmd_mturn_acq_abort()
        if ret > 0:
            return -ret

        for bpm in self._devices:
            bpm.acq_repeat = repeat
            bpm.acq_channel = acq_rate
            bpm.acq_trigger = trig
            bpm.acq_nrsamples_pre = nr_points_before
            bpm.acq_nrsamples_post = nr_points_after

        return self.cmd_mturn_acq_start()

    def cmd_mturn_acq_abort(self, wait=True, timeout=10) -> int:
        """Abort BPMs acquistion.

        Args:
            wait (bool, optional): whether or not to wait BPMs get ready.
                Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for bpm in self._devices:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Abort

        if wait:
            return self.wait_acquisition_finish(timeout=timeout)
        return 0

    def wait_acquisition_finish(self, timeout=10) -> int:
        """Wait for all BPMs to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, bpm in enumerate(self._devices):
            t0_ = _time.time()
            if not bpm.wait_acq_finish(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def cmd_mturn_acq_start(self, wait=True, timeout=10) -> int:
        """Start BPMs acquisition.

        Args:
            wait (bool, optional): whether or not to wait BPMs get ready.
                Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for bpm in self._devices:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Start
        if wait:
            return self.wait_acquisition_start(timeout=timeout)
        return 0

    def wait_acquisition_start(self, timeout=10) -> bool:
        """Wait for all BPMs to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, bpm in enumerate(self._devices):
            t0_ = _time.time()
            if not bpm.wait_acq_start(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def set_switching_mode(self, mode='direct'):
        """Set switching mode of BPMS.

        Args:
            mode ((str, int), optional): Desired mode, must be in
                {'direct', 'switching', 1, 3}. Defaults to 'direct'.

        Raises:
            ValueError: When mode is not in {'direct', 'switching', 1, 3}.

        """
        if mode not in ('direct', 'switching', 1, 3):
            raise ValueError('Value must be in ("direct", "switching", 1, 3).')

        for bpm in self._devices:
            bpm.switching_mode = mode

    def mturn_update_initial_timestamps(self, consider_sum=False):
        """Call this method before acquisition to get orbit for comparison."""
        self._initial_timestamps = self.get_mturn_timestamps(
            return_sum=consider_sum)

    def mturn_reset_flags(self):
        """Reset Multiturn flags to wait for a new orbit update."""
        for flag in self._mturn_flags.values():
            flag.clear()

    def mturn_reset_flags_and_update_initial_timestamps(
            self, consider_sum=False):
        """Set initial state to wait for orbit acquisition to start."""
        self.mturn_reset_flags()
        self.mturn_update_initial_timestamps(consider_sum)

    def mturn_wait_update_flags(self, timeout=10):
        """Wait for all acquisition flags to be updated.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, flag in enumerate(self._mturn_flags.values()):
            t00 = _time.time()
            if not flag.wait(timeout=timeout):
                return i + 1
            timeout -= _time.time() - t00
            timeout = max(timeout, 0)
        return 0

    def mturn_wait_update_timestamps(
            self, timeout=10, consider_sum=False) -> int:
        """Call this method after acquisition to check if data was updated.

        For this method to work it is necessary to call
            mturn_update_initial_timestamps
        before the acquisition starts, so that a reference for comparison is
        created.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.
            consider_sum (bool, optional): Whether to also wait for sum signal
                to be updated. Defaults to False.

        Returns:
            int: code describing what happened:
                -1: initial timestamps were not defined;
                =0: data updated.
                >0: index of the first BPM which did not update plus 1.

        """
        if self._initial_timestamps is None:
            return -1

        tsmp0 = self._initial_timestamps
        while timeout > 0:
            t00 = _time.time()
            tsmp = self.get_mturn_timestamps(return_sum=consider_sum)
            errors = _np.any(_np.equal(tsmp, tsmp0), axis=1)
            if not _np.any(errors):
                return 0
            _time.sleep(0.1)
            timeout -= _time.time() - t00

        return int(_np.nonzero(errors)[0][0])+1

    def mturn_wait_update(self, timeout=10, consider_sum=False) -> int:
        """Combine all methods to wait update data.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.
            consider_sum (bool, optional): Whether to also wait for sum signal
                to be updated. Defaults to False.

        Returns:
            int: code describing what happened:
                -1: initial timestamps were not defined;
                =0: data updated.
                >0: index of the first BPM which did not update plus 1.

        """
        t00 = _time.time()
        ret = self.mturn_wait_update_flags(timeout)
        if ret > 0:
            return ret
        timeout -= _time.time() - t00

        return self.mturn_wait_update_timestamps(
            timeout, consider_sum=consider_sum)

    def _mturn_set_flag(self, pvname, **kwargs):
        _ = kwargs
        self._mturn_flags[pvname].set()



#!/usr/bin/env python-sirius

"""Test type module."""

from unittest import TestCase

from siriuspy.bsmp import Types


class TestBSMPTypes(TestCase):
    """Test BSMP Types."""

    def test_char_type(self):
        """Test char type properties."""
        t = Types.T_CHAR

        self.assertEqual(t.type, 'char')
        self.assertEqual(t.size, 1)
        self.assertEqual(t.fmt, '<c')
        self.assertFalse(t.check(1))
        self.assertFalse(t.check(1.5))
        self.assertTrue(t.check('c'))

    def test_uint8_type(self):
        """Test uint8 type properties."""
        t = Types.T_UINT8

        self.assertEqual(t.type, 'uint8')
        self.assertEqual(t.size, 1)
        self.assertEqual(t.fmt, '<B')
        self.assertFalse(t.check(1.5))
        self.assertFalse(t.check('c'))
        self.assertTrue(t.check(1))

    def test_uint16_type(self):
        """Test uint16 type properties."""
        t = Types.T_UINT16

        self.assertEqual(t.type, 'uint16')
        self.assertEqual(t.size, 2)
        self.assertEqual(t.fmt, '<H')
        self.assertFalse(t.check(1.5))
        self.assertFalse(t.check('c'))
        self.assertTrue(t.check(1))

    def test_uint32_type(self):
        """Test uint32 type properties."""
        t = Types.T_UINT32

        self.assertEqual(t.type, 'uint32')
        self.assertEqual(t.size, 4)
        self.assertEqual(t.fmt, '<I')
        self.assertFalse(t.check(1.5))
        self.assertFalse(t.check('c'))
        self.assertTrue(t.check(1))

    def test_float_type(self):
        """Test float type properties."""
        t = Types.T_FLOAT

        self.assertEqual(t.type, 'float')
        self.assertEqual(t.size, 4)
        self.assertEqual(t.fmt, '<f')
        self.assertTrue(t.check(1.5))
        self.assertFalse(t.check('c'))
        self.assertFalse(t.check(1))
